// ======================================================================
//
// Gl.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Gl.h"

#include "Bitmap.h"
#include "ConfigEngine.h"
#include "ConfigPlatform.h"
#include "ExitChain.h"
#include "Game.h"
#include "gl_init.h"
#include "IoWin.h"
#include "Material.h"
#include "Mouse.h"
#include "Registry.h"
#include "Os.h"

#include <stdio.h>
#include <stdlib.h>

// ======================================================================

typedef void (FAR *InitGl)(const GL_init *);

// ======================================================================

const int DEFAULT_SCREEN_WIDTH  = 640;
const int DEFAULT_SCREEN_HEIGHT = 480;
const int DEFAULT_SCREEN_DEPTH  =   2;

// ======================================================================

int            Gl::screenWidth;
int            Gl::screenHeight;
int            Gl::screenDepthBytes;

int            Gl::materialToken;

ClipInfo      *Gl::clipInfo;
ClipInfo       Gl::clipInfoStack[CLIP_INFO_STACK];

int            Gl::frameSkip;

GL_abstract   *Gl::gl;
int            Gl::rasterMajor;
int            Gl::rasterMinor;
byte           Gl::glRMovieSizeShift;
byte           Gl::glRMoviePlaceShift;
byte           Gl::glGMovieSizeShift;
byte           Gl::glGMoviePlaceShift;
byte           Gl::glBMovieSizeShift;
byte           Gl::glBMoviePlaceShift;

int                   Gl::alternatingBackgroundMaterial;
const StaticMaterial *Gl::backgroundMaterial;
const StaticMaterial *Gl::currentBackgroundMaterial;
const StaticMaterial *Gl::singleFrameBackgroundMaterial;

HINSTANCE      Gl::dllHandle;
int            Gl::dllNumber;
Gl::CreateGl   Gl::createGl;
Gl::DestroyGl  Gl::destroyGl;

bool           Gl::useSuperDetailFlag;
bool           Gl::lightingRGB;
bool           Gl::translucency;
bool           Gl::fog;
bool           Gl::bilinearFiltering;
bool           Gl::lensFlares;

bool           Gl::warningOneMegVideoCard;
bool           Gl::warningBackBufferInSystemRam;
bool           Gl::warningBankSwitched;

bool           Gl::mode2d;

int            Gl::textureMemory;

PrimitiveProcessing    Gl::primitiveProcessing;
const StaticMaterial  *Gl::material;
bool                   Gl::alphaBlendEnabled = false;
AlphaBlendFunction     Gl::alphaBlendSource = ABF_One;
AlphaBlendFunction     Gl::alphaBlendDestination = ABF_Zero;

// ======================================================================

void Gl::install(void)
{
	int newWidth, newHeight, newDepthBytes;

	// setup the clip info
	clipInfo = clipInfoStack;

	// init
	dllHandle = NULL;
	dllNumber = -1;

	// get the initial screen size
	newWidth       = ConfigPlatform::screenWidth();
	newHeight      = ConfigPlatform::screenHeight();
	newDepthBytes  = ConfigPlatform::screenDepth();

	// get the initial rasterizer mode
	rasterMajor    = ConfigPlatform::rasterMajor();
	rasterMinor    = ConfigPlatform::rasterMinor();

	// see if we should look for the DLL to load from the registry
	if (rasterMajor == 0)
	{
		bool      result;
		Registry  reg;
		DWORD     dword, dwordLen;

		dwordLen = sizeof(dword);
	
		result = reg.open(HKEY_LOCAL_MACHINE, Game::getRegistryPath()) && reg.getValue ("Video Card Number", REG_DWORD, &dword, dwordLen);

#if USE_DEFAULT_DLL == 0
		FATAL(!result, ("Could not determine what GL DLL to load"));
		rasterMajor = dword;
#else
		if (result)
			rasterMajor = dword;
		else
			rasterMajor = USE_DEFAULT_DLL;
#endif
	}

	mode2d = true;

	// if no screensize was set, pick the default
	if (!newWidth)
		newWidth  = DEFAULT_SCREEN_WIDTH;
	if (!newHeight)
		newHeight = DEFAULT_SCREEN_HEIGHT;
	if (!newDepthBytes)
		newDepthBytes = DEFAULT_SCREEN_DEPTH;

	ExitChain::add(Gl::remove, "gl", 1);

	primitiveProcessing = PP_PassAll;
	material = NULL;

	// set the screen size	
	resize(newWidth, newHeight, newDepthBytes);

	// keep the 3dfx logo from showing up when task switching back into the game
	SetEnvironmentVariable("fx_glide_no_splash", "1");
}

// ----------------------------------------------------------------------

void Gl::remove(void)
{
	if (!ExitChain::isFataling())
	{
		verifyMode2d();

		// resize the screen down to nothing
		resize(0, 0, 0);
	}
	else
	{
		delete gl;
		gl = NULL;
	}
}

// ----------------------------------------------------------------------

void Gl::unloadDll(void)
{
	verifyMode2d();

	if (dllNumber >= 0)
	{
		if (dllHandle)
			FreeLibrary(dllHandle);
		dllHandle = NULL;
		dllNumber = -1;
		createGl = NULL;
		destroyGl = NULL;
	}
}

// ----------------------------------------------------------------------

void Gl::loadDll(int newNumber)
{
	char name[32];

	verifyMode2d();

	// unload an old dll we may have had
	unloadDll();

	// figure out the name of the dll
	dllNumber = newNumber;	
	sprintf(name, "gl_%05d.dll", dllNumber);

	// load the dll
	MESSAGE(true, ("loading raster dll %s\n", name));
	dllHandle = LoadLibrary(name);
	FATAL(!dllHandle, ("Please re-run the install program.  (Gl::loadDLL load failed %d, %s)", dllNumber, Os::getLastError()));

	// fetch the creation routine
	createGl  = (CreateGl)GetProcAddress(dllHandle, "CreateGl");
	DEBUG_FATAL(!createGl, ("Gl::loadDLL create failed %d", dllNumber));

	// fetch the destruction routine
	destroyGl = (DestroyGl)GetProcAddress(dllHandle, "DestroyGl");
	DEBUG_FATAL(!destroyGl, ("Gl::loadDLL destroy failed %d", dllNumber));
	
	InitGl initGl;
	initGl  = (InitGl)GetProcAddress(dllHandle, "InitGl");
	DEBUG_FATAL(!initGl, ("Gl::loadDLL init failed %d", dllNumber));

	GL_init init;
	memset(&init, 0, sizeof(init));

	// memory allocation routines
	init.MemoryManager__alloc          = MemoryManager::alloc;
	init.MemoryManager__free           = MemoryManager::free;

	init.MemoryManager__getGuardPattern = MemoryManager::getGuardPattern;
	init.MemoryManager__getUserPattern = MemoryManager::getUserPattern;
	init.MemoryManager__getFreePattern = MemoryManager::getFreePattern;

	// fatal routine to call
  init.Fatal                         = Fatal;
  init.Warning                       = Warning;
  init.Message                       = Message;

	// camera render stats
	init.cameraRenderStats             = Camera::renderStats;

	// engine parmaters
	init.engineParmData                = ConfigEngine::data;

	// platform parameters
	init.platformParmData              = ConfigPlatform::data;

	// profile
	PROFILE_GET_DLL(&init.Profile__indx, &init.Profile__entry);

	// dprint
#if USE_DEBUG_PRINT
	init.doNotUseDPrintClear           = DoNotUseDPrintClear;
	init.doNotUseDPrint                = DoNotUseDPrint;
	init.doNotUseDPrintFlag            = DoNotUseDPrintFlag;
	init.doNotUseDPrintPaused          = DoNotUseDPrintPaused;
	init.doNotUseDPrintLogging         = DoNotUseDPrintLogging;
	init.doNotUseDPrintExplicitLogging = DoNotUseDPrintExplicitLogging;
#else
	init.doNotUseDPrintClear           = NULL;
	init.doNotUseDPrint                = NULL;
	init.doNotUseDPrintFlag            = NULL;
	init.doNotUseDPrintPaused          = NULL;
	init.doNotUseDPrintLogging         = NULL;
	init.doNotUseDPrintExplicitLogging = NULL;
#endif

	// dflag
#if USE_DEBUG_FLAG
	init.doNotUseDFlag                 = DoNotUseDFlag;
#else
	init.doNotUseDFlag                 = NULL;
#endif

	// dkey
#if USE_DEBUG_KEY
	init.doNotUseDKeyContext           = DoNotUseDKeyContext;
	init.doNotUseDKeyPressed           = DoNotUseDKeyPressed;
	init.doNotUseDKeyDown              = DoNotUseDKeyDown;
#else
	init.doNotUseDKeyContext           = NULL;
	init.doNotUseDKeyPressed           = NULL;
	init.doNotUseDKeyDown              = NULL;
#endif

	initGl(&init);
}

// ----------------------------------------------------------------------

void Gl::resize(int newWidth, int newHeight, int newDepthBytes, bool force)
{
	GL_input  input;
	GL_output output;
	int       oldWidth, oldHeight, oldDepthBytes;

	verifyMode2d();

	DEBUG_FATAL(clipInfo != clipInfoStack, ("Gl::resize clipping stack not empty"));

	// ignore unforced changes to the same mode
	if (!force && (newWidth == screenWidth && newHeight == screenHeight && newDepthBytes == screenDepthBytes))
		return;

	// save old and set the new screen dimensions
	oldWidth         = screenWidth;
	oldHeight        = screenHeight;
	oldDepthBytes    = screenDepthBytes;
	screenWidth      = newWidth;
	screenHeight     = newHeight;
	screenDepthBytes = newDepthBytes;

	// delete the old virtual gl
	if (gl)
	{
		MaterialList::destroyAllData();
		PaletteList::unloadAllData();

		destroyGl(gl);
		gl = NULL;
	}

	// create the new virtual gl
	if (screenWidth && screenHeight && screenDepthBytes)
	{
		// clear the gl init structs
		memset(&input, 0, sizeof(input));
		memset(&output, 0, sizeof(output));

		// handle to the window
		input.window   = Os::getWindow();

		// setup the virtual gl initialization structure
		input.width    = screenWidth;
		input.height   = screenHeight;
		input.depth    = screenDepthBytes;
		input.clipInfo = const_cast<const ClipInfo **>(&clipInfo); // -qq- why is this const cast needed?
		input.major    = rasterMajor;
		input.minor    = rasterMinor;
		input.mfc      = USE_MFC;

		// make sure we have the correct gl dll loaded
		loadDll(rasterMajor);

		// create the correct gl
		gl = createGl();
		FATAL(!gl, ("Gl::resize create failed"));

		// make sure the gl is usable
		FATAL(!gl->query(rasterMinor), ("Please re-run the install program.  (Gl::resize query failed)"));

		// get the version
		GL_abstract::Information information;
		gl->getInformation(&information);
		MESSAGE(true, ("Gl::resize information %s %d.%d=version %d=ranking\n", information.name, information.versionMajor, information.versionMinor, information.ranking));

		// init the gl
		FATAL(!gl->init(&input, &output), ("Please re-run the install program.  (Gl::resize init failed)"));

		useSuperDetailFlag = output.useSuperDetailFlag;
		lightingRGB = output.lightingRGB;
		translucency = output.translucency;
		fog = output.fog;
		bilinearFiltering = output.bilinearFiltering;
		lensFlares = output.lensFlares;

		warningOneMegVideoCard = output.warningOneMegVideoCard;
		warningBackBufferInSystemRam = output.warningBackBufferInSystemRam;
		warningBankSwitched = output.warningBankSwitched;
		textureMemory = output.textureMemory;

		glRMovieSizeShift  = output.rMovieSizeShift;
		glRMoviePlaceShift = output.rMoviePlaceShift;
		glGMovieSizeShift  = output.gMovieSizeShift;
		glGMoviePlaceShift = output.gMoviePlaceShift;
		glBMovieSizeShift  = output.bMovieSizeShift;
		glBMoviePlaceShift = output.bMoviePlaceShift;

		// reset some state variables in the DLL 
		setPrimitiveProcessing(primitiveProcessing);
		setMaterial(material);
		setAlphaBlendEnabled(alphaBlendEnabled);
		setAlphaBlendFunctions(alphaBlendSource, alphaBlendDestination);

		MaterialList::createAllData();
	}
	else
		unloadDll();

	Mouse::unconstrain();

	// no material downloaded to the dll yet
	materialToken = 0;

	// no background in the dll yet
	currentBackgroundMaterial = NULL;

	// set clipping to fullscreen
	clipInfo->set(0, 0, width() - 1, height() - 1, true);

	Game::resize(oldWidth, oldHeight, oldDepthBytes, newWidth, newHeight, newDepthBytes);
	IoWinManager::resize(oldWidth, oldHeight, oldDepthBytes, newWidth, newHeight, newDepthBytes);
}

// ----------------------------------------------------------------------

void Gl::clipPush(int x0, int y0, int x1, int y1)
{
	clipInfo++;
	DEBUG_FATAL(clipInfo >= clipInfoStack+CLIP_INFO_STACK, ("Gl::clipPush overflow"));
	clipInfo->set(x0, y0, x1, y1, true);
}

// ----------------------------------------------------------------------

void Gl::clipOff(void)
{
	clipInfo++;
	DEBUG_FATAL(clipInfo >= clipInfoStack+CLIP_INFO_STACK, ("Gl::clipOff overflow"));
	clipInfo->set(0, 0, width() - 1, height() - 1, false);
}

// ----------------------------------------------------------------------

void Gl::clipPop(void)
{
	DEBUG_FATAL(clipInfo == clipInfoStack, ("Gl::clipPop underflow"));
	clipInfo--;
}

// ----------------------------------------------------------------------

void Gl::setRasterizer(int major, int minor)
{
	verifyMode2d();

	if (rasterMajor != major || rasterMinor != minor)
	{
		rasterMajor = major;
		rasterMinor = minor;
		resize(width(), height(), depthBytes(), 1);
	}
}

// ----------------------------------------------------------------------

void Gl::lock(void)
{
	bool result;

	verifyMode2d();

	result = gl->lock();
	if (!result)
	{
		// force a resize to reinstall the GL
		DPRINT_LOG(("Gl::lock failed, resizing\n"));
		resize(width(), height(), depthBytes(), true);

		// try the lock a second time
		DPRINT_LOG(("Gl::lock failed, trying second lock\n"));
		result = gl->lock();
		FATAL(!result, ("Gl::lock 2nd fail"));

		DPRINT_LOG(("Gl::lock failed, second lock successful\n"));
	}
}

// ----------------------------------------------------------------------

void Gl::unlock(void)
{
	const StaticMaterial *wantBackgroundMaterial;

	verifyMode2d();

	// setup the alternative flashing RGB clearing
	if (DFlag(DF_E_ALT_CLEAR_RGB))
	{
		switch (alternatingBackgroundMaterial)
		{
			case 0:
				wantBackgroundMaterial = MaterialList::red()->getStaticMaterial();
				alternatingBackgroundMaterial = 1;
				break;

			case 1:
				wantBackgroundMaterial = MaterialList::green()->getStaticMaterial();
				alternatingBackgroundMaterial = 2;
				break;

			case 2:
				wantBackgroundMaterial = MaterialList::blue()->getStaticMaterial();
				alternatingBackgroundMaterial = 0;
				break;

			default:
				DEBUG_FATAL(1, ("Gl::unlock alternate bad index %d", alternatingBackgroundMaterial));
				wantBackgroundMaterial = MaterialList::black()->getStaticMaterial();
		}
	}
	else
		if (singleFrameBackgroundMaterial)
		{
			wantBackgroundMaterial = singleFrameBackgroundMaterial;
			singleFrameBackgroundMaterial = NULL;
		}
		else
			wantBackgroundMaterial = backgroundMaterial;

	if (wantBackgroundMaterial != currentBackgroundMaterial)
	{
		gl->setBackgroundMaterial(wantBackgroundMaterial);
		currentBackgroundMaterial = wantBackgroundMaterial;
	}

	if (frameSkip)
	{
		gl->unlock(0);
		frameSkip--;
	}
	else
		gl->unlock(1);
}

// ----------------------------------------------------------------------

void Gl::point(int x, int y, real z, real ooz)
{
	TVertex v;

	v.x = static_cast<real>(x);
	v.y = static_cast<real>(y);
	v.z = z;
	v.ooz = ooz;

	v.u = CONST_REAL(0);
	v.v = CONST_REAL(0);

	v.r = CONST_REAL(1);
	v.g = CONST_REAL(1);
	v.b = CONST_REAL(1);
	v.a = CONST_REAL(0);

	drawPointList(&v, 1);
}

// ----------------------------------------------------------------------

void Gl::line(int x0, int y0, int x1, int y1, real z, real ooz)
{
	TVertex v[2];

	for (int i = 0; i < 2; ++i)
	{
		v[i].z = z;
		v[i].ooz = ooz;

		v[i].u = CONST_REAL(0);
		v[i].v = CONST_REAL(0);

		v[i].r = CONST_REAL(1);
		v[i].g = CONST_REAL(1);
		v[i].b = CONST_REAL(1);
		v[i].a = CONST_REAL(1);
	}

	v[0].x = static_cast<real>(x0);
	v[0].y = static_cast<real>(y0);

	v[1].x = static_cast<real>(x1);
	v[1].y = static_cast<real>(y1);
	
	drawLineList(v, 2);
}

// ----------------------------------------------------------------------

void Gl::rect(int x0, int y0, int x1, int y1, bool filled, real z, real ooz)
{
	TVertex v[5];

	for (int i = 0; i < 4; ++i)
	{
		v[i].z = z;
		v[i].ooz = ooz;

		v[i].u = CONST_REAL(0);
		v[i].v = CONST_REAL(0);

		v[i].r = CONST_REAL(1);
		v[i].g = CONST_REAL(1);
		v[i].b = CONST_REAL(1);
		v[i].a = CONST_REAL(0);
	}

	v[0].x = static_cast<real>(x0);
	v[0].y = static_cast<real>(y0);

	v[1].x = static_cast<real>(x1);
	v[1].y = static_cast<real>(y0);

	v[2].x = static_cast<real>(x1);
	v[2].y = static_cast<real>(y1);

	v[3].x = static_cast<real>(x0);
	v[3].y = static_cast<real>(y1);

	if (filled)
		drawTriangleFan(v, 4);
	else
	{
		// connect back to the first vertex
		v[4] = v[0];

		drawLineStrip(v, 5);
	}
}

// ----------------------------------------------------------------------

void Gl::point(const Vector &v)
{
	if (setVertexList(&v, 1, NULL, 0))
	{
		UVertex v;

		v.vertexIndex = 0;
		v.normalIndex = -1;

		v.u = CONST_REAL(0);
		v.v = CONST_REAL(0);

		v.r = CONST_REAL(1);
		v.g = CONST_REAL(1);
		v.b = CONST_REAL(1);
		v.a = CONST_REAL(1);

		drawPointList(&v, 1);
	}
}

// ----------------------------------------------------------------------

void Gl::line(const Vector &v0, const Vector &v1)
{
	Vector vert[2];

	vert[0] = v0;
	vert[1] = v1;

	if (setVertexList(vert, 2, NULL, 0))
	{
		UVertex v[2];

		v[0].vertexIndex = 0;
		v[1].vertexIndex = 1;

		for (int i = 0; i < 2; ++i)
		{
			v[0].normalIndex = -1;

			v[i].u = CONST_REAL(0);
			v[i].v = CONST_REAL(0);

			v[i].r = CONST_REAL(1);
			v[i].g = CONST_REAL(1);
			v[i].b = CONST_REAL(1);
			v[i].a = CONST_REAL(1);
		}

		drawLineList(v, 2);
	}
}

// ----------------------------------------------------------------------

void Gl::modeError(void)
{
	int tryMode, inMode;

	if (mode2d)
	{
		tryMode = 3;
		inMode  = 2;
	}
	else
	{
		tryMode = 2;
		inMode  = 3;
	}

	DEBUG_FATAL(true,  ("Gl::modeError %dd function attempted in %dd mode", tryMode, inMode));
}

// ======================================================================
